//
//Example of skeletal animation via R2VB 
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include	"libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include	<stdio.h>
#include	<stdlib.h>

#include	"libTexture.h"
#include	"Vector3D.h"
#include	"Vector2D.h"
#include	"Vector4D.h"
#include	"Data.h"
#include	"GlslProgram.h"
#include	"FrameBuffer.h"
#include	"VertexBuffer.h"
#include	"StringUtils.h"
#include	"md5-model.h"
#include	"utils.h"
#include	"fpTexture.h"

Vector3D	eye   ( 0, 0, 150 );			// camera position
Vector3D	light ( 5, 0, 4 );			// light position
float	 	angle = 0;
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
unsigned	diffMap;

FrameBuffer	buffer  ( 512, 200 );

Vector4D * vecBuf = new Vector4D [ 512*200];
GlslProgram	program;
GLuint		jointsMap;				// texture rect map with joinst data
GLuint		weightsMap;				// texture rect with all wieghts for the model
GLuint		vertexWeightsMap;		// texture rect with weightStart, weightCount for every vertex in model
GLuint		vertexMap;				// texture with vertex coordinates for every vertex

VertexBuffer	* vertexBuffer;		// vertex coordinates
VertexBuffer	* texCoordBuffer;	// vertex attributes array (texCoord)
GLenum			mapFormat;
int				vertexMapHeight;
float         * jointsBuf = NULL;

void reshape ( int w, int h );

unsigned tMap;

/////////////////////////////////////////////////////////////////////////////////

HierarchyItem  * hierarchy = NULL;
BaseFrameJoint * baseFrame = NULL;
float         ** frames    = NULL;
Joint          * joints    = NULL;
int				numJoints  = 0;
int				numFrames  = 0;
int				numComponents = 0;						// numAnimatedComponents
int				frame         = 0;
bool			drawSkeleton  = false;
Vector3D	boxMin ( 0, 0, 0 );			// bounding box for an entire animation
Vector3D	boxMax ( 0, 0, 0 );
Md5Model	model;

void	renormalize ( Quaternion& q )
{
	double	len = 1 - q.x*q.x - q.y*q.y - q.z*q.z;

	if ( len < 1e-8 )
		q.w = 0;
	else
		q.w = -sqrt ( len );
}

void	setNumJoints ( int n )
{
	hierarchy = new HierarchyItem  [numJoints = n];
	baseFrame = new BaseFrameJoint [numJoints];
}
	
void	setNumFrames ( int n )
{
	frames = new float * [numFrames = n];
}

bool	loadHierarchy ( Data * data )
{
	string	str, cmd, args;
	int		parentJointIndex, flags, startIndex;

	for ( int i = 0; ; )
	{
		if ( !getLine ( data, str ) )
			break;

		parseString ( str, cmd, args );

		if ( cmd.empty () )
			continue;

		if ( cmd == "}" )
			return true;

		sscanf ( args.c_str (), "%d %d %d", &parentJointIndex, &flags, &startIndex );

		if ( i < numJoints )
		{
			hierarchy [i].name       = cmd;
			hierarchy [i].parent     = parentJointIndex;
			hierarchy [i].flags      = flags;
			hierarchy [i].startIndex = startIndex;

			i++;
		}
	}

	return true;
}

bool	loadBounds ( Data * data )
{
	string		str, cmd, args;
	Vector3D	vMin, vMax;

	for ( int i = 0; ; )
	{
		if ( !getLine ( data, str ) )
			break;

		parseString ( str, cmd, args );

		if ( cmd.empty () )
			continue;

		if ( cmd == "}" )
			return true;

		if ( cmd == "(" )
			args = cmd + args;

		sscanf ( args.c_str (), "( %f %f %f ) ( %f %f %f )", &vMin.x, &vMin.y, &vMin.z, &vMax.x, &vMax.y, &vMax.z );

		if ( i < numFrames )
		{
			boxMin = Vector3D :: vmin ( boxMin, vMin );
			boxMax = Vector3D :: vmax ( boxMax, vMax );
			i++;
		}
	}

	return true;
}

bool	loadBaseFrame ( Data * data )
{
	string		str, cmd, args;
	Vector3D	pos, rot;

	for ( int i = 0; ; )
	{
		if ( !getLine ( data, str ) )
			break;

		str = trim ( str );

		if ( str == "}" )
			return true;

		sscanf ( str.c_str (), "( %f %f %f ) ( %f %f %f )", &pos.x, &pos.y, &pos.z, &rot.x, &rot.y, &rot.z );

		baseFrame [i].pos    = pos;
		baseFrame [i].orient = Quaternion ( rot.x, rot.y, rot.z, 0 );

		renormalize ( baseFrame [i++].orient );
	}

	return true;
}

bool	loadFrame ( Data * data, int no )
{
	string	str, cmd, args;

	frames [no] = new float [numComponents];

	for ( int i = 0; ; )
	{
		if ( !getLine ( data, str ) )
			break;

		do
		{
			parseString ( str, cmd, args );

			if ( cmd.empty () )
				break;

			if ( cmd == "}" )
				return true;

			sscanf ( cmd.c_str (), "%f", &frames [no][i++] );

			str = args;
		} while ( args != "" );
	}

	return true;
}

void	buildJoints ()
{
	for ( int i = 0; i < numJoints; i++ )
		if ( joints [i].parent != NULL )
		{
			joints [i].pos    = joints [i].parent -> pos + joints [i].parent -> orient.rotate ( joints [i].pos );
			joints [i].orient = joints [i].parent -> orient * joints [i].orient;
		}
}

void	resetJoints ()
{
	for ( int i = 0; i < numJoints; i++ )
	{
		joints [i].name        = hierarchy [i].name;
		joints [i].parentIndex = hierarchy [i].parent;
		joints [i].parent      = joints [i].parentIndex == -1 ? NULL : &joints [joints[i].parentIndex];
		joints [i].pos         = baseFrame [i].pos;
		joints [i].orient      = baseFrame [i].orient;
	}
}

bool	load ( const string& fileName )
{
	Data * data = getFile ( fileName.c_str () );

	printf ( "Loading animation %s\n", fileName.c_str () );

	if ( data == NULL )
	{
		printf ( "Animation %s not found.\n", fileName.c_str () );

		return false;
	}

	string	str, cmd, args;

	for ( ; ; )
	{
		if ( !getLine ( data, str ) )
			break;

		parseString ( str, cmd, args );

		if ( cmd.empty () )
			continue;

		if ( cmd == "MD5Version" )
		{

		}
		else
		if ( cmd == "commandline" )
		{

		}
		else
		if ( cmd == "numFrames" )
			setNumFrames ( atoi ( args.c_str () ) );
		else
		if ( cmd == "numJoints" )
			setNumJoints ( atoi ( args.c_str () ) );
		else
		if ( cmd == "frameRate" )
			;//setFrameRate ( atoi ( args.c_str () ) );
		else
		if ( cmd == "numAnimatedComponents" )
			numComponents = atoi ( args.c_str () );
		else
		if ( cmd == "hierarchy" )
			loadHierarchy ( data );
		else
		if ( cmd == "bounds" )
			loadBounds ( data );
		else
		if ( cmd == "baseframe" )
			loadBaseFrame ( data );
		else
		if ( cmd == "frame" )
			loadFrame ( data, atoi ( args.c_str () ) );
	}

	delete data;

	printf ( "Preparing joints\n" );

	joints = new Joint [numJoints];
	
	resetJoints ();
	buildJoints ();
	
	return true;
}

void	setFrame ( int no )
{
	if ( no < 0 )
		no = numFrames - 1;
		
	if ( no >= numFrames )
		no = 0;
	
	frame = no;
	
	resetJoints ();
	
	for ( int i = 0; i < numJoints; i++ )
	{
		int	flags = hierarchy [i].flags;
		int	pos   = hierarchy [i].startIndex;
		
		if ( flags & 1 )
			joints [i].pos.x = frames [no][pos++];
			
		if ( flags & 2 )
			joints [i].pos.y = frames [no][pos++];
			
		if ( flags & 4 )
			joints [i].pos.z = frames [no][pos++];
			
		if ( flags & 8 )
			joints [i].orient.x = frames [no][pos++];
			
		if ( flags & 16 )
			joints [i].orient.y = frames [no][pos++];
			
		if ( flags & 32 )
			joints [i].orient.z = frames [no][pos++];
		
		renormalize ( joints [i].orient );
	}
	
	buildJoints ();
	
//	model.compute ( joints );
	
	printf ( "Frame:%d\n", frame );
}

bool	createJointsMap ()
{
    glGenTextures   ( 1, &jointsMap );
    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, jointsMap );
    glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 );                         // set 1-byte alignment

    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP );    // set texture to repeat mode
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	
    glTexImage2D    ( GL_TEXTURE_RECTANGLE_ARB, 0, mapFormat, 2*numJoints, 1, 0, 
	                  GL_RGBA, GL_FLOAT, NULL );
    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, 0 );

	GLenum error = glGetError ();

	printf  ( "createJointsMap%d:%s\n", error, gluErrorString ( error ) );

	return true;
}

bool	loadJointsMap ()
{
	if ( jointsBuf == NULL )
		jointsBuf = (float *) calloc ( 1, numJoints * 8 * sizeof ( float ) );
		
	if ( jointsBuf == NULL )
		return false;
		
								// now remap joints array to buffer
	for ( int i = 0; i < numJoints; i++ )
	{
		jointsBuf [i*8    ] = joints [i].pos.x;
		jointsBuf [i*8 + 1] = joints [i].pos.y;
		jointsBuf [i*8 + 2] = joints [i].pos.z;
		jointsBuf [i*8 + 3] = joints [i].parentIndex;
		jointsBuf [i*8 + 4] = joints [i].orient.x;
		jointsBuf [i*8 + 5] = joints [i].orient.y;
		jointsBuf [i*8 + 6] = joints [i].orient.z;
		jointsBuf [i*8 + 7] = joints [i].orient.w;
	}
	
    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, jointsMap );
    glTexImage2D    ( GL_TEXTURE_RECTANGLE_ARB, 0, mapFormat, 2*numJoints, 1, 0, 
	                  GL_RGBA, GL_FLOAT, jointsBuf );
    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, 0 );

	return true;
}

//
// Build using model vertexWeightsMap and weightsMap
//

bool	buildWeightAndVertexMaps ()
{
	int	weightsCount = 0;
	int	vertexCount  = 0;
	int	i;
	
	for ( i = 0; i < model.numMeshes; i++ )
	{
		Md5Mesh& mesh = model.meshes [i];
		
										// fix vertex ## in tris list
		for ( int j = 0; j < mesh.numTris; j++ )
			for ( int k = 0; k < 3; k++ )
				mesh.tris [j].index [k] += vertexCount;

		for ( int l = 0; l < mesh.numVertices; l++ )
				mesh.vertices [l].weightIndex += weightsCount;

		weightsCount += mesh.numWeights;
		vertexCount  += mesh.numVertices;
	}
										// we need weightsCount*2 texels in weightsMap
	int	width         = 512;			// use this width for all our fixed maps
	int	weightsHeight = (weightsCount * 2 + width - 1 ) / width;
	int	vertexHeight  = (vertexCount + width - 1) / width;
	int	count         = 0;

										// now prepare data to initialize maps
	float * weightBuf = (float *) calloc ( 4*4, width * weightsHeight );
	
	vertexMapHeight = vertexHeight;
	
	for ( i = 0; i < model.numMeshes; i++ )
	{
		Md5Mesh& mesh = model.meshes [i];
		
		mesh.weightsStart = count;
		
		for ( int j = 0; j < mesh.numWeights; j++ )
		{
			weightBuf [count++] = mesh.weights [j].pos.x;
			weightBuf [count++] = mesh.weights [j].pos.y;
			weightBuf [count++] = mesh.weights [j].pos.z;
			weightBuf [count++] = mesh.weights [j].weight;
			weightBuf [count++] = mesh.weights [j].jointIndex;
			weightBuf [count++] = 0;
			weightBuf [count++] = 0;
			weightBuf [count++] = 0;
		}
	}
	
	glEnable        ( GL_TEXTURE_RECTANGLE_ARB );
    glGenTextures   ( 1, &weightsMap );
    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, weightsMap );
    glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 ); 

    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S,     GL_CLAMP ); 
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T,     GL_CLAMP );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	
    glTexImage2D    ( GL_TEXTURE_RECTANGLE_ARB, 0, mapFormat, width, weightsHeight, 0, 
	                  GL_RGBA, GL_FLOAT, weightBuf );
	GLenum error = glGetError ();

	printf  ( "%d:%s\n", error, gluErrorString ( error ) );

    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, 0 );
	
	free ( weightBuf );
										
	float * vertexBuf = (float *) calloc ( 4*4, width * vertexHeight );

	for ( i = count = 0; i < model.numMeshes; i++ )
	{
		Md5Mesh& mesh = model.meshes [i];
		
		mesh.vertexStart = count;
		
		for ( int j = 0; j < mesh.numVertices; j++ )
		{
			vertexBuf [count++] = mesh.vertices [j].weightIndex;
			vertexBuf [count++] = mesh.vertices [j].weightCount;
			vertexBuf [count++] = 0;
			vertexBuf [count++] = 0;
		}
	}

    glGenTextures   ( 1, &vertexWeightsMap );
    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, vertexWeightsMap );
    glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 ); 

    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S,     GL_CLAMP );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T,     GL_CLAMP );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	
    glTexImage2D    ( GL_TEXTURE_RECTANGLE_ARB, 0, mapFormat, width, vertexHeight, 0, 
	                  GL_RGBA, GL_FLOAT, vertexBuf );

	error = glGetError ();

	printf  ( "%d:%s\n", error, gluErrorString ( error ) );

    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, 0 );
	
	free ( vertexBuf );
	
	return true;
}

void	createTexCoordBuffer ()
{
										// create and fill buffer for texture coordinates
	float * buf = (float *) malloc ( buffer.getWidth () * buffer.getHeight () * 2 
	                                 * sizeof ( float ) );
	int		i, count;
	
	for ( i = count = 0; i < model.numMeshes; i++ )
	{
		Md5Mesh& mesh = model.meshes [i];
		
		for ( int j = 0; j < mesh.numVertices; j++ )
		{
			buf [count++] = mesh.vertices [j].tex.x;
			buf [count++] = mesh.vertices [j].tex.y;
		}
	}
	
	texCoordBuffer = new VertexBuffer ();
	
	texCoordBuffer -> bind    ( GL_ARRAY_BUFFER_ARB );
	texCoordBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 2 * sizeof ( float ), 
	                            buf, GL_STATIC_DRAW );
	texCoordBuffer -> unbind  ();
	
	free ( buf );
}	

void	createVertexBuffer ()
{
	vertexBuffer = new VertexBuffer ();
	
	vertexBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	vertexBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 4 * sizeof ( float ), 
	                          NULL, GL_DYNAMIC_DRAW );
	vertexBuffer -> unbind  ();
}

void	drawRect ( int w, int h )
{
	glEnable      ( GL_TEXTURE_RECTANGLE_ARB );
    glBegin       ( GL_QUADS );
        glTexCoord2f ( 0, 0 );
        glVertex2f   ( 0, 0 );

        glTexCoord2f ( w, 0 );
        glVertex2f   ( w, 0 );

        glTexCoord2f ( w, h );
        glVertex2f   ( w, h );

        glTexCoord2f ( 0, h );
        glVertex2f   ( 0, h );
    glEnd   ();
}


//
// create vertex coordinates using r2vb
//
void	buildVertices ()
{
	loadJointsMap ();									// load current joints into texture


	glEnable ( GL_TEXTURE_RECTANGLE_ARB );
	glDisable ( GL_TEXTURE_2D );

	glActiveTextureARB ( GL_TEXTURE1_ARB );
    glBindTexture      ( GL_TEXTURE_RECTANGLE_ARB, weightsMap );

	glActiveTextureARB ( GL_TEXTURE2_ARB );
    glBindTexture      ( GL_TEXTURE_RECTANGLE_ARB, jointsMap );
	
	glActiveTextureARB ( GL_TEXTURE0_ARB );
    glBindTexture      ( GL_TEXTURE_RECTANGLE_ARB, vertexWeightsMap );


	program.bind   ();
	buffer.bind    ();
	startOrtho     ( 512, 200 );
	drawRect       ( 512, 200 /*vertexMapHeight*/ );
	endOrtho       ();
	program.unbind ();
	
												// now copy vertex data from texture to buffer
	vertexBuffer -> bind   ( GL_PIXEL_PACK_BUFFER_ARB );
	glBindTexture          ( GL_TEXTURE_RECTANGLE_ARB, buffer.getColorBuffer () );
	glReadPixels           ( 0, 0, buffer.getWidth (), 200 /*vertexMapHeight*/, GL_RGBA, 
	                         GL_FLOAT, NULL );
	buffer.unbind          ();
	vertexBuffer -> unbind ();

}

////////////////////////////////////////////////////////////////////////////////////

void display ()
{
	buildVertices ();
	
	glClear      ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix ();

	Vector3D	len   = boxMax - boxMin;
	float		scale = 0.38f;	//100.0f / len.length ();
	
	glTranslatef ( -eye.x, -eye.y, -eye.z );
    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );
	glScalef     ( scale, scale, scale );


    glPushClientAttrib   ( GL_CLIENT_VERTEX_ARRAY_BIT );
	glEnableClientState  ( GL_VERTEX_ARRAY );
	glEnableClientState  ( GL_TEXTURE_COORD_ARRAY );

	vertexBuffer -> bind ( GL_ARRAY_BUFFER_ARB	);
	glVertexPointer ( 4, GL_FLOAT, 0, NULL );

	texCoordBuffer -> bind ( GL_ARRAY_BUFFER_ARB );
	glTexCoordPointer   ( 2, GL_FLOAT, 0, NULL );
	

	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glDisable          ( GL_TEXTURE_RECTANGLE_ARB );
	glEnable           ( GL_TEXTURE_2D );

	for ( int i = 0; i < model.numMeshes; i++ )
	{
		Md5Mesh& mesh = model.meshes [i];
		
		glBindTexture      ( GL_TEXTURE_2D, mesh.diffuseMap );
		
                                            // request draw
		glDrawElements ( GL_TRIANGLES, 3*mesh.numTris, GL_UNSIGNED_INT, &mesh.tris [0] );
	}
	
	vertexBuffer   -> unbind ();
	texCoordBuffer -> unbind ();

	glPopClientAttrib ();
	glPopMatrix     ();

	glutSwapBuffers ();
}

void reshape ( int w, int h )
{
	glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
	glMatrixMode   ( GL_PROJECTION );
	glLoadIdentity ();
	gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1, 5000.0 );
	glMatrixMode   ( GL_MODELVIEW );
	glLoadIdentity ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
    	exit ( 0 );
	else
	if ( key == '+' )
		setFrame ( frame + 1 );
	else
	if ( key == '-' )
		setFrame ( frame - 1 );
	else
	if ( key == ' ' )
		drawSkeleton = !drawSkeleton;
}

void	animate ()
{
	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( 512, 512 );

								// create window
	glutCreateWindow ( "Skeletal animation via R2VB" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
	glutIdleFunc     ( animate );

	init                      ();
	initExtensions            ();
	assertExtensionsSupported ( "ARB_pixel_buffer_object ARB_texture_rectangle" );

	if ( (mapFormat = fpFormatWithPrecision ( 32 ) ) == 0 )
	{
		printf ( "Floating-point textures not supported !\n" );
		
		return 1;
	}
	
	if ( !FrameBuffer :: isSupported () )
	{
		printf ( "Fbo not supported\n" );
		
		return 1;
	}
	
	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported\n" );
		
		return 1;
	}
	
	unsigned tempMap = buffer.createColorRectTexture ( GL_RGBA, mapFormat );
	
	buffer.create ();
	buffer.bind   ();
	if ( !buffer.attachColorTexture ( GL_TEXTURE_RECTANGLE_ARB, tempMap ) )
		printf ( "buffer error with color attachment\n");

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );
					
	buffer.unbind ();

    if ( !program.loadShaders ( "skeletal.vsh", "skeletal.fsh" ) )
    {
        printf ( "Error loading skeletal animation shaders:\n%s\n", program.getLog ().c_str () );

        return 3;
    }
	
	string	doomPath  = DOOM_PATH;
	string	modelName = "maggot3";
	string	animName  = "walk.md5anim";

	if ( argc > 1 )
		modelName =  argv [1];
		
	if ( argc > 2 )
		animName = string ( argv [2] ) + ".md5anim";
		
	if ( argc > 3 )
		doomPath = string ( argv [3] ) + "/";
		
	string	pak0 = doomPath + "pak000.pk4";
	string	pak1 = doomPath + "pak001.pk4";
	string	pak2 = doomPath + "pak002.pk4";
	string	pak3 = doomPath + "pak003.pk4";
	string	pak4 = doomPath + "pak004.pk4";

	printf           ( "loading model %s.\n", modelName.c_str () );
	addZipFileSystem ( pak2.c_str () );
	addZipFileSystem ( pak4.c_str () );
		
	model.load    ( modelName );
										// now load the animation
	string	fileName  = string  ( "models/md5/monsters/" ) + modelName + "/" + animName;
		
	if ( !load ( fileName.c_str () ) )
	{
		printf ( "Error loading anim file \"%s\".\n", modelName.c_str () );
		
		return 1;
	}
	
	model.compute ( joints );
	
	printf ( "Model loaded.\n" );

	if ( !createJointsMap () )
	{
		printf ( "Error in createJointsMap.\n" );
		
		return 2;
	}

	if ( !buildWeightAndVertexMaps () )
	{
		printf ( "Error in buildWeightAndVertexMaps.\n" );
		
		return 3;
	}

	createVertexBuffer   ();
	createTexCoordBuffer ();

	program.bind       ();
	program.setTexture ( "vertDataMap", 0 );
	program.setTexture ( "weightsMap",  1 );
	program.setTexture ( "jointsMap",   2 );
	program.unbind     ();
	

    glutMainLoop ();

	return 0;
}
