//
// Example of rendering noise-based water using render-to-vertex-buffer in OpenGL,
//
// Author: Alex V. Boreskoff <steps3d@gmail.co>, <steps3d@narod.ru>
//

#include	"libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include	<stdio.h>
#include	<stdlib.h>

#include	"libTexture.h"
#include	"libTexture3D.h"
#include	"Vector3D.h"
#include	"Vector2D.h"
#include	"Vector4D.h"
#include	"Data.h"
#include	"GlslProgram.h"
#include	"FrameBuffer.h"
#include	"VertexBuffer.h"
#include	"utils.h"
#include	"fpTexture.h"

#define	TEX_SIZE	512				// width & height of texture

Vector3D	eye   ( 7, 5, 7 );		// camera position
Vector3D	light ( 5, 0, 4 );		// light position
float	 	angle = 0;
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
float		waveScale = 0.2;

GLenum		mapFormat;				// texture format used for vertex map
unsigned	noiseMap;				// map with source image
unsigned	reflectionMap;			// cubic map with reflection
unsigned	verticesMap;			// map with resulting vertices
unsigned	normalsMap;				// map with resulting normals

FrameBuffer	buffer  ( TEX_SIZE, TEX_SIZE );
GlslProgram	program;
GlslProgram	drawProgram;
unsigned	jointsMap;				// texture rect map with joinst data
unsigned	weightsMap;				// texture rect with all wieghts for the model
unsigned	vertexWeightsMap;		// texture rect with weightStart, weightCount for every vertex in model
unsigned	vertexMap;				// texture with vertex coordinates for every vertex
bool		wireframe = false;

VertexBuffer	* vertexBuffer;		// vertex coordinates
VertexBuffer    * normalBuffer;		// normal buffer
VertexBuffer	* indexBuffer;		// vertex index buffer

void reshape ( int w, int h );

void	createIndexBuffer ()
{
	int * data = (int *) malloc ( (TEX_SIZE-1)*(TEX_SIZE-1)*6*sizeof (int) );
	int	  k    = 0;
	
	for ( int i = 0; i < TEX_SIZE - 1; i++ )
		for ( int j = 0; j < TEX_SIZE - 1; j++ )
		{
			data [k]   = i + TEX_SIZE*j;				// first triangle (i,j)-(i+1,j) - (i,j+1)
			data [k+1] = i + TEX_SIZE*j + 1;
			data [k+2] = i + TEX_SIZE*j + TEX_SIZE;
			
			data [k+3] = i + TEX_SIZE*j + 1;			// second triangle (i+1,j)-(i+1,j+1) - (i,j+1)
			data [k+4] = i + TEX_SIZE*j + TEX_SIZE + 1;
			data [k+5] = i + TEX_SIZE*j + TEX_SIZE;
			
			k         += 6;
		}
		
	indexBuffer = new VertexBuffer ();
	
	indexBuffer -> bind    ( GL_ELEMENT_ARRAY_BUFFER_ARB );
	indexBuffer -> setData ( (TEX_SIZE - 1) * (TEX_SIZE - 1) * 6 * sizeof ( int ), data, GL_STATIC_DRAW );
	indexBuffer -> unbind  ();
	
	free ( data );
}

void	createVertexBuffer ()
{
	vertexBuffer = new VertexBuffer ();
	
	vertexBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	vertexBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 4 * sizeof ( float ), NULL, GL_DYNAMIC_DRAW );
	vertexBuffer -> unbind  ();
}

void	createNormalBuffer ()
{
	normalBuffer = new VertexBuffer ();
	
	normalBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	normalBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 4 * sizeof ( float ), NULL, GL_DYNAMIC_DRAW );
	normalBuffer -> unbind  ();
}

//
// create vertex coordinates using r2vb
//
void	buildVertices ()
{
	GLenum	buffers [] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };

														// load source image map
	glActiveTextureARB ( GL_TEXTURE0_ARB );
    glBindTexture      ( GL_TEXTURE_3D, noiseMap );

	program.bind   ();									// perform rendering to FBO texture
	buffer.bind    ();
	glDrawBuffers  ( 2, buffers );
	startOrtho     ( TEX_SIZE, TEX_SIZE );
	drawQuad       ( TEX_SIZE, TEX_SIZE );
	endOrtho       ();
	program.unbind ();
	
														// copy vertex data from texture to vertex buffer
	glReadBuffer           ( GL_COLOR_ATTACHMENT0_EXT );
	vertexBuffer -> bind   ( GL_PIXEL_PACK_BUFFER_ARB );
//	glBindTexture          ( GL_TEXTURE_RECTANGLE_ARB, buffer.getColorBuffer ( 0 ) );
	glReadPixels           ( 0, 0, TEX_SIZE, TEX_SIZE, GL_RGBA, GL_FLOAT, NULL );
	vertexBuffer -> unbind ();

	glReadBuffer           ( GL_COLOR_ATTACHMENT1_EXT );
	normalBuffer -> bind   ( GL_PIXEL_PACK_BUFFER_ARB );
//	glBindTexture          ( GL_TEXTURE_RECTANGLE_ARB, buffer.getColorBuffer ( 1 ) );
	glReadPixels           ( 0, 0, TEX_SIZE, TEX_SIZE, GL_RGBA, GL_FLOAT, NULL );
	buffer.unbind          ();
	normalBuffer -> unbind ();
	
}

void display ()
{
	buildVertices ();
	
	glClear      ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix ();

	float		scale = 1.0f;
	
    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );
	glScalef     ( scale, scale, 2 );
	glTranslatef ( -10, -10, -1 );

	glActiveTextureARB ( GL_TEXTURE0_ARB );
    glBindTexture      ( GL_TEXTURE_3D, noiseMap );

	glActiveTextureARB ( GL_TEXTURE1_ARB );
    glBindTexture      ( GL_TEXTURE_CUBE_MAP, reflectionMap );
	
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glColor3f          ( 1, 1, 1 );
	
    glPushClientAttrib   ( GL_CLIENT_VERTEX_ARRAY_BIT );
	
	glEnableClientState  ( GL_VERTEX_ARRAY );
	vertexBuffer -> bind ( GL_ARRAY_BUFFER_ARB );
	glVertexPointer      ( 4, GL_FLOAT, 0, NULL );
	
	glEnableClientState  ( GL_TEXTURE_COORD_ARRAY );
	normalBuffer -> bind ( GL_ARRAY_BUFFER_ARB );
	glTexCoordPointer    ( 4, GL_FLOAT, 0, NULL );
	
	indexBuffer  -> bind ( GL_ELEMENT_ARRAY_BUFFER_ARB );
    glIndexPointer       ( GL_UNSIGNED_INT, 0, 0 );
	
	if ( wireframe )
		glPolygonMode ( GL_FRONT_AND_BACK, GL_LINE );
		
	drawProgram.bind   ();
	glDrawElements     ( GL_TRIANGLES, (TEX_SIZE-1)*(TEX_SIZE-1)*6, GL_UNSIGNED_INT, 0 );
	drawProgram.unbind ();

	glPolygonMode     ( GL_FRONT_AND_BACK, GL_FILL );	// restore polygon mode
	glPopClientAttrib ();
	glPopMatrix       ();
	glutSwapBuffers   ();
}

void reshape ( int w, int h )
{
	glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
	glMatrixMode   ( GL_PROJECTION );
	glLoadIdentity ();
	gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1, 5000.0 );
	gluLookAt      ( eye.x, eye.y, eye.z,	// eye
					 0, 0, 0,				// center
					 0, 0, 1 );				// up


	glMatrixMode   ( GL_MODELVIEW );
	glLoadIdentity ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
    	exit ( 0 );
		
	if ( key == '+' )
		waveScale += 0.01;
	else
	if ( key == '-' )
		waveScale -= 0.01;
		
	if ( key == ' ' )
		wireframe = !wireframe;
}

void	animate ()
{
	float	time = 0.001f * glutGet ( GLUT_ELAPSED_TIME );
	
	program.bind ();
	program.setUniformFloat ( "time",      time );
	program.setUniformFloat ( "waveScale", waveScale );
	program.unbind ();

	drawProgram.bind ();
	drawProgram.setUniformFloat  ( "time", time );
	drawProgram.setUniformVector ( "eye",  eye  );
	drawProgram.unbind ();
	
	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( 512, 512 );

								// create window
	glutCreateWindow ( "Example of rendering water via render-to-vertex-buffer" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
	glutIdleFunc     ( animate );

	init                      ();
	initExtensions            ();
	assertExtensionsSupported ( "ARB_pixel_buffer_object" );

	if ( (mapFormat = fpFormatWithPrecision ( 32 ) ) == 0 )
	{
		printf ( "Floating-point textures not supported !\n" );
		
		return 1;
	}
	
	if ( !FrameBuffer :: isSupported () )
	{
		printf ( "Fbo not supported\n" );
		
		return 1;
	}
	
	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported\n" );
		
		return 1;
	}
	
	unsigned vertMap = buffer.createColorRectTexture ( GL_RGBA, mapFormat );
	unsigned normMap = buffer.createColorRectTexture ( GL_RGBA, mapFormat );
	
	buffer.create ();
	buffer.bind   ();
	if ( !buffer.attachColorTexture ( GL_TEXTURE_RECTANGLE_ARB, vertMap, 0 ) )
		printf ( "buffer error with color attachment 0\n");

	if ( !buffer.attachColorTexture ( GL_TEXTURE_RECTANGLE_ARB, normMap, 1 ) )
		printf ( "buffer error with color attachment 1\n");

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );
					
	buffer.unbind ();

    if ( !program.loadShaders ( "r2vb-water.vsh", "r2vb-water.fsh" ) )
    {
        printf ( "Error loading r2vb-water shaders:\n%s\n", program.getLog ().c_str () );

        return 3;
    }

    if ( !drawProgram.loadShaders ( "water-draw.vsh", "water-draw.fsh" ) )
    {
        printf ( "Error loading water-draw shaders:\n%s\n", drawProgram.getLog ().c_str () );

        return 4;
    }

	program.bind       ();
	program.setTexture ( "noiseMap", 0 );
	program.unbind     ();
	
	drawProgram.bind       ();
	drawProgram.setTexture ( "noiseMap",      0 );
	drawProgram.setTexture ( "reflectionMap", 1 );
	drawProgram.unbind     ();
	
	const char * faces [6] =
	{
		"../../Textures/Cubemaps/skyrt.bmp",
		"../../Textures/Cubemaps/skylf.bmp",
		"../../Textures/Cubemaps/skydn.bmp",
		"../../Textures/Cubemaps/skyup.bmp",
		"../../Textures/Cubemaps/skybk.bmp",
		"../../Textures/Cubemaps/skyfr.bmp",
	};

	noiseMap      = createTexture3D ( false, "../../Textures/Noise/noise-3D.dds" );
	reflectionMap = createCubeMap   ( true, faces );
	
	if ( noiseMap == 0 )
	{
		printf ( "Texture noise-3D.dds not found.\n" );
		
		return 1;
	}
	
	createVertexBuffer ();
	createNormalBuffer ();
	createIndexBuffer  ();
	
	printf ( "Use space to toggle normal/wireframe mode.\nUse mouse to rotate mesh.\n" );
	
    glutMainLoop ();

	return 0;
}
