//
// Example of render-to-vertex-buffer in OpenGL,
// Based on Apple's example 
//
// Author: Alex V. Boreskoff <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include	"libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include	<stdio.h>
#include	<stdlib.h>

#include	"libTexture.h"
#include	"Vector3D.h"
#include	"Vector2D.h"
#include	"Vector4D.h"
#include	"Data.h"
#include	"GlslProgram.h"
#include	"FrameBuffer.h"
#include	"VertexBuffer.h"
#include	"utils.h"
#include	"fpTexture.h"

#define	TEX_SIZE	512				// width & height of texture

Vector3D	eye   ( 7, 5, 7 );		// camera position
Vector3D	light ( 5, 0, 4 );		// light position
float	 	angle = 0;
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;

GLenum		mapFormat;				// texture format used for vertex map
unsigned	srcMap;					// map with source image
unsigned	verticesMap;			// map with resulting vertices

FrameBuffer	buffer  ( TEX_SIZE, TEX_SIZE );
GlslProgram	program;
GlslProgram	drawProgram;
unsigned	jointsMap;				// texture rect map with joinst data
unsigned	weightsMap;				// texture rect with all wieghts for the model
unsigned	vertexWeightsMap;		// texture rect with weightStart, weightCount for every vertex in model
unsigned	vertexMap;				// texture with vertex coordinates for every vertex
bool		wireframe = false;

VertexBuffer	* vertexBuffer;		// vertex coordinates
VertexBuffer	* indexBuffer;		// vertex index buffer

void reshape ( int w, int h );

void	createIndexBuffer ()
{
	int * data = (int *) malloc ( (TEX_SIZE-1)*(TEX_SIZE-1)*6*sizeof (int) );
	int	  k    = 0;
	
	for ( int i = 0; i < TEX_SIZE - 1; i++ )
		for ( int j = 0; j < TEX_SIZE - 1; j++ )
		{
			data [k]   = i + TEX_SIZE*j;				// first triangle (i,j)-(i+1,j) - (i,j+1)
			data [k+1] = i + TEX_SIZE*j + 1;
			data [k+2] = i + TEX_SIZE*j + TEX_SIZE;
			
			data [k+3] = i + TEX_SIZE*j + 1;			// second triangle (i+1,j)-(i+1,j+1) - (i,j+1)
			data [k+4] = i + TEX_SIZE*j + TEX_SIZE + 1;
			data [k+5] = i + TEX_SIZE*j + TEX_SIZE;
			
			k         += 6;
		}
		
	indexBuffer = new VertexBuffer ();
	
	indexBuffer -> bind    ( GL_ELEMENT_ARRAY_BUFFER_ARB );
	indexBuffer -> setData ( (TEX_SIZE - 1) * (TEX_SIZE - 1) * 6 * sizeof ( int ), data, GL_STATIC_DRAW );
	indexBuffer -> unbind  ();
	
	free ( data );
}

void	createVertexBuffer ()
{
	vertexBuffer = new VertexBuffer ();
	
	vertexBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	vertexBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 4 * sizeof ( float ), NULL, GL_DYNAMIC_DRAW );
	vertexBuffer -> unbind  ();
}

//
// create vertex coordinates using r2vb
//
void	buildVertices ()
{
														// load source image map
	glActiveTextureARB ( GL_TEXTURE0_ARB );
    glBindTexture      ( GL_TEXTURE_2D, srcMap );

	program.bind  ();									// perform rendering to FBO texture
	buffer.bind   ();
	startOrtho    ( TEX_SIZE, TEX_SIZE );
	drawQuad      ( TEX_SIZE, TEX_SIZE );
	endOrtho      ();
	program.unbind ();
	
														// copy vertex data from texture to vertex buffer
	vertexBuffer -> bind   ( GL_PIXEL_PACK_BUFFER_ARB );
//	glBindTexture          ( GL_TEXTURE_RECTANGLE_ARB, buffer.getColorBuffer () );
	glReadPixels           ( 0, 0, TEX_SIZE, TEX_SIZE, GL_RGBA, GL_FLOAT, NULL );
	buffer.unbind          ();
	vertexBuffer -> unbind ();
}

void display ()
{
	buildVertices ();
	
	glClear      ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix ();

	float		scale = 1.0f;
	
    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );
	glScalef     ( scale, scale, 2 );
	glTranslatef ( -5, -5, -1 );

	glColor3f ( 1, 1, 1 );
	
    glPushClientAttrib  ( GL_CLIENT_VERTEX_ARRAY_BIT );
	
	glEnableClientState  ( GL_VERTEX_ARRAY );
	vertexBuffer -> bind ( GL_ARRAY_BUFFER );
	glVertexPointer      ( 4, GL_FLOAT, 0, NULL );
	
	indexBuffer  -> bind ( GL_ELEMENT_ARRAY_BUFFER_ARB );
    glIndexPointer       ( GL_UNSIGNED_INT, 0, 0 );
	
	if ( wireframe )
		glPolygonMode ( GL_FRONT_AND_BACK, GL_LINE );
		
	drawProgram.bind   ();
	glDrawElements     ( GL_TRIANGLES, (TEX_SIZE-1)*(TEX_SIZE-1)*6, GL_UNSIGNED_INT, 0 );
	drawProgram.unbind ();

	glPolygonMode     ( GL_FRONT_AND_BACK, GL_FILL );	// restore polygon mode
	glPopClientAttrib ();
	glPopMatrix       ();
	glutSwapBuffers   ();
}

void reshape ( int w, int h )
{
	glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
	glMatrixMode   ( GL_PROJECTION );
	glLoadIdentity ();
	gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1, 5000.0 );
	gluLookAt      ( eye.x, eye.y, eye.z,	// eye
					 0, 0, 0,				// center
					 0, 0, 1 );				// up


	glMatrixMode   ( GL_MODELVIEW );
	glLoadIdentity ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
    	exit ( 0 );
		
	if ( key == ' ' )
		wireframe = !wireframe;
}

void	animate ()
{
	float	time = 0.001f * glutGet ( GLUT_ELAPSED_TIME );
	
	program.bind ();
	program.setUniformFloat ( "time", time );
	program.unbind ();
	
	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( 512, 512 );

								// create window
	glutCreateWindow ( "Example of rendering to vertex buffer" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
	glutIdleFunc     ( animate );

	init                      ();
	initExtensions            ();
	assertExtensionsSupported ( "ARB_pixel_buffer_object" );

	if ( (mapFormat = fpFormatWithPrecision ( 32 ) ) == 0 )
	{
		printf ( "Floating-point textures not supported !\n" );
		
		return 1;
	}
	
	if ( !FrameBuffer :: isSupported () )
	{
		printf ( "Fbo not supported\n" );
		
		return 1;
	}
	
	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported\n" );
		
		return 1;
	}
	
	srcMap = createTexture2D ( false, "../../Textures/wall.tga" );
	
	if ( srcMap == 0 )
	{
		printf ( "Texture wall.tga not found.\n" );
		
		return 1;
	}
	
	unsigned tempMap = buffer.createColorRectTexture ( GL_RGBA, mapFormat );
	
	buffer.create ();
	buffer.bind   ();
	if ( !buffer.attachColorTexture ( GL_TEXTURE_RECTANGLE_ARB, tempMap ) )
		printf ( "buffer error with color attachment\n");

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );
					
	buffer.unbind ();

    if ( !program.loadShaders ( "r2vb.vsh", "r2vb.fsh" ) )
    {
        printf ( "Error loading r2vb shaders:\n%s\n", program.getLog ().c_str () );

        return 3;
    }
	
    if ( !drawProgram.loadShaders ( "r2vb-draw.vsh", "r2vb-draw.fsh" ) )
    {
        printf ( "Error loading r2vb-draw shaders:\n%s\n", drawProgram.getLog ().c_str () );

        return 4;
    }
	
	program.bind       ();
	program.setTexture ( "srcMap", 0 );
	program.unbind     ();
	
	createVertexBuffer ();
	createIndexBuffer  ();
	
	printf ( "Use space to toggle normal/wireframe mode.\nUse mouse to rotate mesh.\n" );
	
    glutMainLoop ();

	return 0;
}
